import importlib
import json
import logging
import traceback
from abc import abstractmethod

from django.conf import settings
from django.urls import resolve
from rest_framework import status
from rest_framework.test import APITestCase, APIRequestFactory, APIClient

from drftest import doc_generator
from drftest.abc_test_meta import ABCTestMeta
from drftest.auth_provider import AuthProvider
from drftest.uuid_encoder import UUIDEncoder

ABCTestMeta.add_ignored_test_class_name('BaseViewTest')


class BaseViewTest(APITestCase, metaclass=ABCTestMeta):
    auth_provider_class = None
    current_test_name = None
    current_test_doc = None

    def setUp(self):
        super(BaseViewTest, self).setUp()
        self.api_client = APIClient()
        self.request_factory = APIRequestFactory()

    def _get_auth_provider(self) -> AuthProvider:
        if not self.auth_provider_class:
            try:
                module_name, class_name = settings.DRF_TEST_AUTH_PROVIDER_CLASS.rsplit(".", 1)
            except AttributeError as e:
                msg = 'Value of DRF_TEST_AUTH_PROVIDER_CLASS should be set in settings'
                raise AttributeError(msg) from e
            self.auth_provider_class = getattr(importlib.import_module(module_name), class_name)
        return self.auth_provider_class()

    @abstractmethod
    def _make_url(self, kwargs=None):
        """
        :param kwargs: path parameters in a form which can get passed to django's
        reverse function
        :return: url which is being tested. (Most probably generated using django's reverse
        function)
        """
        raise NotImplementedError()

    def test_resolves_view(self, *_):
        self.assertIsNotNone(resolve(self._make_url(self._get_default_url_kwargs())))

    def _get_default_url_kwargs(self):
        """
        :return: A dictionary where the keys are name of path parameters and the values
        are their default values. It doesn't matter if the values are not meaningful
        defaults. The output is just used by tests generated by drftest which check
        whether or not things like url mapping, etc are working and your view is working.
        """
        return None

    def test_calling_endpoint(self, *_):
        try:
            response = self.api_client.get(self._make_url(self._get_default_url_kwargs()))
            self.assertNotEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        except Exception:
            traceback.print_exc()
            self.fail("Something caused request to endpoint to raise an exception.")

    def assertStatus(self, response, expected):
        msg = 'Expected status code of response to equal {status}'.format(status=expected)
        if hasattr(response, 'data'):
            msg += '\nResponse data is {data}'.format(data=response.data)
        self.assertEqual(response.status_code, expected, msg)

    def assertSuccess(self, response, expected_status_code=None):
        data_appendix = ''
        if hasattr(response, 'data'):
            data_appendix = 'Response data is {data}'.format(data=response.data)
        if expected_status_code is None:
            self.assertIn(
                response.status_code, [
                    status.HTTP_200_OK,
                    status.HTTP_201_CREATED,
                ],
                """
                200 or 201 status code should get returned
                {response_data}
                """.format(response_data=data_appendix)
            )
        else:
            self.assertEqual(
                response.status_code, expected_status_code,
                'Status could should have been {code}'.format(
                    code=expected_status_code) + data_appendix
            )

    def _head_for_response(self, user=None, data=None, url_kwargs=None,
                           docs=True, **extra):
        return self.__request_for_response(self.api_client.head, user=user, data=data, docs=docs,
                                           url_kwargs=url_kwargs, format='json', extra=extra)

    def _get_for_response(self, user=None, data=None, url_kwargs=None, docs=True, **extra):
        return self.__request_for_response(self.api_client.get, user=user, data=data,
                                           url_kwargs=url_kwargs, format='json', docs=docs,
                                           extra=extra)

    def _post_for_response(self, user=None, data=None, url_kwargs=None, format='json',
                           docs=True, **extra):
        return self.__request_for_response(self.api_client.post, user=user, data=data,
                                           url_kwargs=url_kwargs, format=format, docs=docs,
                                           extra=extra)

    def _patch_for_response(self, user=None, data=None, url_kwargs=None, format='json',
                            docs=True, **extra):
        return self.__request_for_response(self.api_client.patch, user=user, data=data,
                                           url_kwargs=url_kwargs, format=format, docs=docs,
                                           extra=extra)

    def _put_for_response(self, user=None, data=None, url_kwargs=None, format='json',
                          docs=True, **extra):
        return self.__request_for_response(self.api_client.put, user=user, data=data,
                                           url_kwargs=url_kwargs, format=format, docs=docs,
                                           extra=extra)

    def _delete_for_response(self, user=None, data=None, url_kwargs=None, format='json', docs=True,
                             **extra):
        return self.__request_for_response(self.api_client.delete, user=user, data=data,
                                           url_kwargs=url_kwargs, format=format,
                                           docs=docs, extra=extra)

    def __request_for_response(self, method, user=None, data=None, url_kwargs=None, format='json',
                               docs=True, extra=None):
        if not docs or format != 'json':
            return
        extra = extra or {}
        self._get_auth_provider().set_auth(self.api_client, user)
        headers = self._modify_headers(extra)
        if method == self.api_client.get:
            response = method(self._make_url(kwargs=url_kwargs), data=data, **headers)
        else:
            response = method(
                self._make_url(kwargs=url_kwargs), data=data, format=format, **headers)
        if format != 'json':
            return response
        self._generate_docs(response, method, data, url_kwargs, format, headers, user)
        return response

    def _ensure_json_serializable(self, obj, fail_silently=True):
        try:
            json.dumps(obj, cls=UUIDEncoder)
            return obj
        except Exception:
            logging.error('**********')
            logging.error('Error happened in {}'.format(self.current_test_name))
            logging.error('**********')
            traceback.print_exc()
            if not fail_silently:
                raise
            return None

    def _modify_headers(self, headers):
        return {
            'HTTP_' + k.upper().replace('-', '_'): v for k, v in headers.items()
        }

    def _generate_docs(self, response, method, data, url_kwargs, format, headers, user):
        if self.current_test_name in ['test_has_permission_classes',
                                      'test_resolves_view',
                                      'test_calling_endpoint']:
            return
        app_name = self.__class__.__module__.split('.')[0]
        url = self._make_url(url_kwargs)
        method_name = {
            self.api_client.get: 'get',
            self.api_client.post: 'post',
            self.api_client.delete: 'delete',
            self.api_client.patch: 'patch',
            self.api_client.put: 'put'
        }[method]
        response_data = response.data if hasattr(response, 'data') else {}
        headers = headers or {}
        headers.update(self._get_auth_provider().get_auth_headers(user))
        doc_generator.class_docs[self.__class__.__name__] = self.__class__.__doc__
        doc_generator.store.append({
            'method': method_name,
            'data': self._ensure_json_serializable(data),
            'url': url,
            'url_kwargs': self._ensure_json_serializable(url_kwargs),
            'format': format,
            'headers': headers,
            'success': 200 <= response.status_code < 300,
            'meta': {
                'docs': self.current_test_doc,
                'method_name': self.current_test_name,
                'class_name': self.__class__.__name__,
                'app_name': app_name
            },
            'response': {
                'data': self._ensure_json_serializable(response_data),
                'status': response.status_code,
            }
        })

    @abstractmethod
    def _get_view_class(self):
        """
        :return: The view class which is being tested. For example if MyAwsomeView is being
        tested, this method should look like `return MyAwsomeView`
        """
        raise NotImplementedError

    def __getattribute__(self, item):
        result = super(BaseViewTest, self).__getattribute__(item)
        if item and callable(result) and item.startswith('test'):
            BaseViewTest.current_test_doc = result.__doc__
            BaseViewTest.current_test_name = result.__name__
        return result



from abc import ABCMeta


class ABCTestMeta(ABCMeta):
    ignored_test_class_names = []

    @classmethod
    def add_ignored_test_class_name(mcs, non_test_class):
        mcs.ignored_test_class_names.append(non_test_class)

    def __new__(mcs, name, bases, namespace):
        cls = super(ABCTestMeta, mcs).__new__(mcs, name, bases, namespace)
        cls.__test__ = cls.__name__ not in mcs.ignored_test_class_names
        return cls



import traceback

from django_nose.runner import NoseTestSuiteRunner

from drftest.doc_generator import write_docs


class TestRunner(NoseTestSuiteRunner):
    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        result = None
        try:
            result = super().run_tests(test_labels, extra_tests)
            write_docs()
        except Exception:
            traceback.print_exc()
        finally:
            return result



from abc import abstractmethod

from rest_framework.test import APIClient


class AuthProvider:
    @abstractmethod
    def set_auth(self, api_client: APIClient, user):
        """
        Authenticates user using the given `api_client`
        """
        raise NotImplementedError()

    def get_auth_headers(self, user):
        """
        Returns a dictionary indicating what headers need to be set for authentication
        purpose of given user.
        It is only used in docs.
        """
        return {}



import json

from django import template
from django.utils.safestring import mark_safe

from drftest.uuid_encoder import UUIDEncoder

register = template.Library()


@register.filter
def to_json(value):
    return mark_safe(json.dumps(value, indent=4, sort_keys=True, cls=UUIDEncoder))



from abc import abstractmethod
from inspect import isclass

from rest_framework.settings import api_settings

from drftest.abc_test_meta import ABCTestMeta
from drftest.base_view_test import BaseViewTest

ABCTestMeta.add_ignored_test_class_name('BaseAuthenticatedViewTest')


class BaseAuthenticatedViewTest(BaseViewTest, metaclass=ABCTestMeta):
    @abstractmethod
    def _get_permission_classes(self):
        """
        Should return a list of permission classes. drftest generates a testcase to check if
        result of calling this method matches those specified by `permission_classes` attribute
        of your class. This way you can test your permission classes separately and in other
        tests just use drftest's auto-generated testcases to check if correct permissions are set.
        """
        raise NotImplementedError()

    def _get_permission_classes_for_docs(self):
        return [str(c) if not isclass(c) else c.__name__
                for c in self._get_permission_classes()]

    @staticmethod
    def __diff(first, second):
        return set(first) - set(second)

    def test_has_permission_classes(self, *_):
        if not self._get_permission_classes():
            self.assertEqual(self._get_view_class().permission_classes,
                             api_settings.DEFAULT_PERMISSION_CLASSES,
                             'Extra permissions are set for the view.')
            return
        xor = set(self._get_permission_classes()) ^ set(self._get_view_class().permission_classes)
        if not xor:
            return
        self.fail('Expected permission classes: {expected} But found: {actual}'.format(
            expected=self._get_permission_classes(),
            actual=self._get_view_class().permission_classes
        ))



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='drftest',
    version='1.0.1',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='drftest is a minimal testing library that aims to facilitate writing tests for '
                'django rest framework views. It also [optionally] generates good looking API '
                'documentations based on tests it runs.',
    install_requires=[
        'django>=1.11.0',
        'djangorestframework',
        'django_nose',
    ],
    long_description=README,
    url='https://github.com/pdrum/drftest',
    author='Pedram Hajesmaeeli',
    author_email='pedram.esmaeeli@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



import os
import shutil
import textwrap
from typing import Dict

from django.conf import settings
from django.template import loader
from django.utils.safestring import mark_safe

"""
{
    'method': 'POST',
    'data': {'a': 'b'},
    'url': '/api/somewhere',
    'url_kwargs': {'a': 'b'},
    'format': 'json',
    'headers': {'a': 'b'},
    'success': 200 <= {response_status_cod} < 300,
    'meta': {
        'docs': '{docstring_for_the_test}',
        'method_name': '{Name of method being tested}',
        'class_name': '{Name of class containing the test}',
        'app_name': 'Name of django app containing the test'
    },
    'response': {
        'data': {'a': 'b'},
        'status': {response status code. eg. 200},
    }
}
"""
store = []
"""
class_docs keeps a dictionary which maps name of test class to its docs.
"""
class_docs = {}


def _categorize_store() -> Dict:
    """
    doc entry for each test case in `BaseViewTest` is a flat dictionary. with no categorization
    based on app, class or method.

    This function categorizes items in that list by (first) app_name and (secondly) class_name.
    """
    categorized = {}
    for test_result in store:
        app_name = test_result['meta']['app_name']
        class_name = test_result['meta']['class_name']
        if app_name not in categorized:
            categorized[app_name] = {}
        if class_name not in categorized[app_name]:
            categorized[app_name][class_name] = {
                'tests': [],
                'description': textwrap.dedent(mark_safe(class_docs.get(class_name) or '')),
            }
        categorized[app_name][class_name]['tests'].append(test_result)
    return categorized


def _get_root_dir():
    """
    mkdocs has a yml file and a docs directory containing yml files. Both of these two
    reside in a directory indicated by settings.DRF_TEST_DOCS_DIR which will be referred to
    as root directory.
    """
    return os.path.expanduser(settings.DRF_TEST_DOCS_DIR)


def _get_docs_path():
    return os.path.join(_get_root_dir(), 'docs')


def _clear_docs_path():
    """
    Removes everything in docs directory in order to prepare it for new docs.
    """
    docs_path = _get_docs_path()
    if os.path.exists(docs_path):
        if os.path.isdir(docs_path):
            shutil.rmtree(docs_path)
        else:
            os.remove(docs_path)
    os.mkdir(docs_path)


def _rewrite_yml(root_dir: str):
    yml_path = os.path.join(root_dir, 'mkdocs.yml')
    with open(yml_path, 'w+') as yml_file:
        yml_file.write('site_name: DRF Tests')


def write_docs():
    if not hasattr(settings, 'DRF_TEST_DOCS_DIR') or not settings.DRF_TEST_DOCS_DIR:
        return

    if not os.path.exists(_get_root_dir()):
        os.mkdir(_get_root_dir())

    _clear_docs_path()
    for app_name, app_docs in _categorize_store().items():
        md_path = os.path.join(_get_docs_path(), '{}.md'.format(app_name))
        t = loader.get_template('doc_of_app.md')
        rendered = t.render({
            'app_name': app_name,
            'app_docs': app_docs,
        })
        with open(md_path, 'w+') as md_file:
            md_file.write(rendered)
    index_path = os.path.join(_get_docs_path(), 'index.md')
    with open(index_path, 'w+') as index_file:
        index_file.write("""
# Welcome

Welcome to *DRF Test* documentation. 

This documentation is categorized according to subsystems of your code.
Name of each subsystem is given in the navigation menu. You can click on 
each item to visit documentation of that subsystem.

Happy coding :)
        """)

    _rewrite_yml(_get_root_dir())
